/* args.c
 *
 * WMAppl - Window Maker Application Launcher
 * Version:  0.5
 * Copyright (C) 1999-2000 Casey Harkins
 *
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "args.h"

/* some globals */
char *rcfile;
IconPath *iconpath;
int debugmode;

/* parse_args
 *  PARAMETERS:
 *      int argc          : number of command line parameters
 *      char **argv       : array of command line parameters
 *  RETURN:
 *      none
 */
void
parse_args(int argc, char **argv)
{
	int c, slen;
	char *home;		/* user's home directory, ending in slash */
	char *tmpiconpath;
	
	/* store user's home directory */
	home = (char *)malloc(sizeof(char)*(strlen((char *)getenv("HOME"))+1));
	strcpy(home,getenv("HOME"));
	if(!home)
	{
		fprintf(stderr, "args.c (warning) : getenv(\"HOME\")\n");
		home = (char *) malloc(sizeof(char)*2);
		strcpy(home, "/");
	}
	slen=strlen(home);
	if(home[slen-1]!='/')
	{
		char *oldhome = home;
		home = realloc(oldhome, (sizeof(char)*(slen+2)));
		if(!home)
		{
			fprintf(stderr, "args.c (error) : realloc()\n");
			exit(1);
		}
		home[slen]='/';
		home[slen+1]=0;
	}

	slen = strlen(home) + 1;

	/* set default options */
	debugmode = 0;
	rcfile = (char *)malloc(sizeof(char)*(slen+strlen(WMAPPLRC)));
	if(!rcfile)
	{
		fprintf(stderr, "args.c (error) : malloc()\n");
		exit(1);
	}
	strcpy(rcfile, home);
	strcat(rcfile, WMAPPLRC);
	tmpiconpath = (char *)malloc(sizeof(char)*(strlen(ICONPATH)+1));
	if(!tmpiconpath)
	{
		fprintf(stderr, "args.c (error) : malloc()\n");
		exit(1);
	}
	strcpy(tmpiconpath, ICONPATH);


	while (1)
	{
		int option_index = 0;

		/* setup long options structure */
		static struct option long_options[] =
		{
			{"rcfile",1,0,'f'},
			{"iconpath",1,0,'i'},
			{"debug",0,0,'d'},
			{"version",0,0,'V'},
			{"help",0,0,'h'},
			{0,0,0,0}
		};
		
		/* get command line option */	
		c = getopt_long(argc, argv, "f:i:dVh", long_options, &option_index);
		
		if(c == -1) break;

		switch(c)
		{
			case 'f':
				if(rcfile) free(rcfile);
				rcfile = (char *)malloc(sizeof(char)*(strlen(optarg)+1));
				if(!rcfile)
				{
					fprintf(stderr, "args.c (error) : malloc()\n");
					exit(1);
				}
				strcpy(rcfile, optarg);
				break;
	
			case 'i':
				if(tmpiconpath) free(tmpiconpath);
				tmpiconpath = (char *)malloc(sizeof(char)*(strlen(optarg)+1));
				if(!tmpiconpath)
				{
					fprintf(stderr, "args.c (error) : malloc()\n");
					exit(1);
				}
				strcpy(tmpiconpath, optarg);
				break;

			case 'd':
				debugmode=1;
				break;

			case 'V':
				show_version();
				break;

			case 'h':
				show_usage();
				break;

			default:
				show_usage();
				break;

		} /* end switch */

	} /* end while */

	/* create the iconpath structure and free the tmpiconpath */
	iconpath = create_iconpath(tmpiconpath);
	free(tmpiconpath);

} /* end parse args */

/* show_version
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
show_version()
{
	printf("\nWMAppl version 0.6\n");
	printf("Copyright (c) 2000 Casey Harkins (charkins@pobox.com)\n");
	exit(1);
} /* end show version */

/* show_usage
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
show_usage()
{
	printf("\nUsage:\n");
	printf("  wmappl [OPTIONS]\n");
	printf("\n");
	printf("  -f, --rcfile [arg]  \tabsolute path of configuration file\n");
	printf("  -i, --iconpath [arg]\tpath to search for icons (paths separated by :'s)\n");
	printf("  -d, --debug         \tdump debug information to stderr\n");
	printf("  -V, --version       \tshow version information\n");
	printf("  -h, --help          \tshow usage information (this info)\n");
	printf("\n");
	printf("Defaults:\n");
	printf("  rcfile=%s\n  iconpath=%s\n",rcfile,ICONPATH);
	if(debugmode) {
		printf("  Debug enabled.\n");
	}
	exit(1);
} /* end show usage */

/* get_rcfile
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      char *  : resource filename, should not return NULL
 */
char *
get_rcfile()
{
	return rcfile;
}

/* get_iconpath
 * 
 *  RETURN:
 *      IconPath *  : the icon path structure 
 */
IconPath *
get_iconpath()
{
	return iconpath;
}

/* create_iconpath
 *  PARAMETERS:
 *      char * : colon separated path
 *  RETURN:
 *      IconPath *  : the icon path as a structure 
 */
IconPath *
create_iconpath(char *colonpath)
{

    int i, j, k;
    char *directory;
    IconPath *return_path;

    /* if path is empty, return NULL */
    if(strlen(colonpath)<1) return (IconPath *)NULL;

    /* get path count */
    for(i=0, j=1; i<strlen(colonpath); i++) {
        if(colonpath[i]==PATHSEP) j++;
    }

    /* allocate new IconPath structure, set count and allocate path array */
    return_path = (IconPath *)malloc(sizeof(IconPath));
    return_path->count = j;
    return_path->path=(char **)malloc(sizeof(char *)*j);

    /* set dirallocsize to the length of the path + 2 (for trailing / and NULL) */
    return_path->dirallocsize = sizeof(char)*(strlen(colonpath)+2);

    /* allocate a string for directory */
    directory = (char *)malloc(return_path->dirallocsize);

    /* iterate through path separating directories */
    for(i=0, j=0, k=0; i<strlen(colonpath); i++, j++) {

        /* are we at a path separator */
        if(colonpath[i]==PATHSEP) {

            /* do we need to add the trailing directory separator */
            if(i>0 && directory[j-1]!=DIRSEP) {
                directory[j]=DIRSEP;
                j++;
            }

            /* terminate the directory with a NULL */
            directory[j]='\0';

            /* reallocate directory to it's proper size */
            directory = (char *)realloc(directory, sizeof(char)*(strlen(directory)+1));
            if(!directory) {
                fprintf(stderr, "wmappl.c: Couldn't reallocate string.\n");
                exit(1);
            }

            /* store the directory in the IconPath structure */
            return_path->path[k] = directory;

            /* allocate new directory string and manage counters */
            directory = (char *)malloc(return_path->dirallocsize);
            k++; j=-1;

        }
        /* if not at path separator, add character to directory string */
        else {
            directory[j]=colonpath[i];
        }
    }
	
	/* do we need to add the trailing directory separator */
	if(i>0 && directory[j-1]!=DIRSEP) {
		directory[j]=DIRSEP;
		j++;
	}

    /* terminate the last directory */
    directory[j]='\0';

    /* reallocate directory to it's proper size */
    directory = (char *)realloc(directory, sizeof(char)*(strlen(directory)+1));
    if(!directory) {
        fprintf(stderr, "wmappl.c: Couldn't reallocate string.\n");
        exit(1);
    }

    /* store the directory in the IconPath structure */
    return_path->path[k]=directory;

    return return_path;
}


/* get_debugmode
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      int     : one if debug option was set, else zero
 */
int
get_debugmode()
{
	return debugmode;
}

