/* dockapp.c - support functions for a WindowMaker docked application
 *  
 * Version: 0.2 
 * Copyright (C) 2000 Casey Harkins (charkins@pobox.com)
 *
 * Authors:
 * 	Casey Harkins (charkins@pobox.com)
 * 	Jarek (talen@charybda.icm.edu.pl)
 * 
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "dockapp.h"

/* create_dockapp
 *	PARAMETERS:
 *		char * appname    : name of application to be registered with X
 *		char * geometry   : X geometry string to use when creating the window
 *		char **interface  : the XPM to use for the background and shape mask
 *      int withdrawn     : if withdrawn is > 0 then the app is drawn in
 *                          withdrawn state (WindowMaker DockApps)
 *                          if withdrawn is zero then it is drawn as a
 *                          normal window
 *		int argc          : number of command line parameters
 *		char **argv		  : array of command line parameters
 *	RETURN:
 *		int : returns non-zero on success
 */
int
create_dockapp(	char *appname,
				char *geometry,
				char **interface,
				int withdrawn,
				int argc,
				char **argv)
{
	/* some X structs */
	XSizeHints xsh;
	XWMHints *xwmh;
	XClassHint xch;
	XGCValues xgcv;
	XTextProperty xtp;
	Window win=0, iconwin=0, rootwin=0;
	Pixmap pixmap, pixmask;
	XpmAttributes pixattr;
	int screen, win_x, win_y, win_w, win_h, garbage;
	unsigned long blackpixel, whitepixel;

	paint_requested = 0;

    /* connect to default display */
    display = XOpenDisplay((char *)getenv("DISPLAY"));
    if(!display)
	{
        fprintf(stderr, "dockapp.c (error) : Couldn't connect to display.\n");
        return(0);
    }

    /* get screen and root window */
    screen = DefaultScreen(display);
    rootwin = RootWindow(display, screen);

	/* get black and white pixel values */
	blackpixel = BlackPixel(display, screen);
	whitepixel = WhitePixel(display, screen);

	/* set Window Manager size hints */
    xsh.flags = USSize | USPosition;
	xsh.width = 64;
	xsh.height = 64;
    XWMGeometry(display, screen, geometry, "64x64+0+0", 0,
        &xsh, &win_x, &win_y, &win_w, &win_h, &garbage);
	
	/* create the normal window */
	win = XCreateSimpleWindow(display, rootwin,
		win_x, win_y, win_w, win_h, 0, blackpixel, whitepixel);
    if(!win)
	{
        fprintf(stderr, 
			"dockapp.c (error): Couldn't create normal window.\n");
        return(0);
    }

	/* set the draw window to the normal window */
	draw_window = win;

	/* if we're using withdrawn state, create the icon window */
	if(withdrawn)
	{
		/* create the icon window */
		iconwin = XCreateSimpleWindow(display, rootwin,
			win_x, win_y, win_w, win_h, 0, blackpixel, whitepixel);
    	if(!iconwin)
		{
        	fprintf(stderr, 
				"dockapp.c (error): Couldn't create icon window.\n");
        	return(0);
    	}
		
		/* set the draw window to the icon window */
		draw_window = iconwin;
	}

	/* set the valuemask on the XpmAttributes structure */
	pixattr.valuemask = 0;
	
	/* load interface pixmap */
	if(XpmCreatePixmapFromData(display, rootwin, interface,
		&pixmap, &pixmask, &pixattr)!=XpmSuccess)
	{
		fprintf(stderr, 
			"dockapp.c (error): Couldn't create interface pixmap.\n");
		return(0);
	}

	/* set background pixmap for the draw window */
	XSetWindowBackgroundPixmap(display, draw_window, pixmap);
	
	/* shape the draw window if supported */
	if(XShapeQueryExtension(display, &garbage, &garbage))
	{
		XShapeCombineMask(display, draw_window, 
			ShapeBounding, 0, 0, pixmask, ShapeSet);
	}
	else
	{
		fprintf(stderr,
			"dockapp.c (warning): Shaped window extension not supported.\n");
	}

	/* setup Window Manager hints */
    xwmh = XAllocWMHints();
    xwmh->flags = InputHint | WindowGroupHint;
    xwmh->input = True;
   	xwmh->window_group = win;

	/* add withdrawn state and icon window hints if we're using withdrawn */
	if(withdrawn)
    {
    	xwmh->flags = xwmh->flags | IconWindowHint | StateHint;
    	xwmh->icon_window = iconwin;
   		xwmh->initial_state = WithdrawnState;
	}
    XSetWMHints(display, win, xwmh);

	/* set class hints */
    xch.res_name = appname;
    xch.res_class = appname;
    XSetClassHint(display, win, &xch);

    /* set size hints */
    XSetWMNormalHints(display, win, &xsh);

    /* tell window manager app name */
    if(!XStringListToTextProperty(&appname, 1, &xtp)) 
	{
        fprintf(stderr, "dockapp.c (error) : Couldn't create text property.\n");
        return(0);
    }
    XSetWMName(display, win, &xtp);

	/* create a graphics context */
    gc = XCreateGC(display, draw_window, GCForeground | GCBackground, &xgcv);
    if(!gc)
	{
        fprintf(stderr, 
			"dockapp.c (error) : Couldn't create graphics context.\n");
        return(0);
    }

	/* select events to catch */
    XSelectInput(display, draw_window,
        ExposureMask |
        ButtonPressMask |
        ButtonReleaseMask |
        StructureNotifyMask);

    /* set the command line for restarting */
    XSetCommand(display, win, argv, argc);

    /* map the main window */
    XMapWindow(display, win);

	return 1;
} /* end create_dockapp */

/* run_dockapp
 *	PARAMETERS:
 *		none
 *	RETURN:
 *		int : returns non-zero on success
 */
int
run_dockapp()
{
	int close=0;
	XEvent e;

	while(!close)
	{

		/* sleep while there are no events pending */
	    while(XPending(display)==0) usleep(_DELAY_TIME);

		/* process all pending events */
	    while(XPending(display))
		{

			/* get next event ifrom queue */
			XNextEvent(display, &e);

			/* handle events */
			switch(e.type)
			{
				case ClientMessage:
				  	paint_requested=1;
				  	break;
				case ButtonPress:
					if(mouse_ptr) mouse_ptr(e.xbutton.x, e.xbutton.y,
						e.xbutton.button, MOUSE_PRESSED);
					break;
				case ButtonRelease:
					if(mouse_ptr) mouse_ptr(e.xbutton.x, e.xbutton.y,
						e.xbutton.button, MOUSE_RELEASED);
					break;
				case Expose:
					if(e.xexpose.count!=0) break;
					paint_requested=1;
					break;
				case ConfigureNotify:
					paint_requested=1;
					break;
				case DestroyNotify:
					close=1;
					break;
			} /* switch */
			/* call paint function if it has been requested */
			if(paint_requested)
			{
				paint_requested = 0;
				if(paint_ptr) paint_ptr(display, draw_window, gc);
			}
			
	    } /* while XPending */
	} /* while not close */

	return 1;
} /* end run_dockapp */

/* request_paint - requests that the paint function be called
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
request_paint()
{
	paint_requested = 1;
}

/* clear_window - clears the draw window
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
clear_window()
{
	XClearWindow(display, draw_window);
}

/* set_paint_func
 *  PARAMETERS:
 *      void (*func)(Display dsp, Drawable drw, GC g) : pointer to paint 
 *                                                      function
 *  RETURN:
 *      int : returns non-zero on success
 */
int
set_paint_func( void (*func)(Display *dsp, Drawable drw, GC g))
{
	if(func)
	{
		paint_ptr=func;
		return 1;
	}
	return 0;
} /* end set_paint_func */

/* set_mouse_func
 *	PARAMETERS:
 *		void (*func)(int x, int y, int b, int s) : pointer to mouse function
 *	RETURN:
 *		int : returns non-zero on success
 */
int
set_mouse_func(	void (*func)(int x, int y, int b, int s))
{
	if(func)
	{
		mouse_ptr=func;
		return 1;
	}
	return 0;
} /* end mouse_func */

/* get_display()
 *  RETURN:
 *      Display * : returns display
 */
Display*
get_display()
{
	return display;
}

/* get_screen()
 *  RETURN:
 *      int : returns screen
 */
int
get_screen()
{
	return DefaultScreen(display);
}

/* get_black_pixel()
 *  RETURN:
 *      unsigned long : black pixel value
 */
unsigned long
get_black_pixel()
{
	return BlackPixel(display, get_screen());

}

/* get_white_pixel()
 *  RETURN:
 *      unsigned long : white pixel value
 */
unsigned long
get_white_pixel()
{
	return WhitePixel(display, get_screen());

}

