/* dockapp.h - support functions for a WindowMaker docked application
 *
 * Version: 0.2
 * Copyright (C) 2000 Casey Harkins (charkins@pobox.com)
 *
 * DESCRIPTION:
 *   This small library of functions should simplify the creation of
 *   WindowMaker docked applications. First the user must make a call to
 *   create_dockapp. Then calls to set_paint_func and set_mouse_func to
 *   set the appropriate callback functions. Finally calling run_dockapp
 *   enters the event loop which calls the appropriate callbacks. The
 *   callbacks receive parameters which will be useful in handling those
 *   events. If more information is needed, the get_XXXXXX calls provide
 *   some additional information. The clear_window and request_paint
 *   functions may also be called from the callbacks to clear the draw
 *   window or request that the paint callback be called.
 *
 *   For information or for updates to these libraries check:
 *       http://www.pobox.com/~charkins/dockapp.html
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef __DOCKAPP_H__

#include <Xlib.h>
#include <Xutil.h>
#include <extensions/shape.h>
#include <xpm.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


//time to wait till the next XPending
#define _DELAY_TIME	100000

/* MOUSE MACROS */
#define MOUSE_PRESSED  1
#define MOUSE_RELEASED 0

/* some globals */
Display *display;
Window draw_window;
GC gc;
int paint_requested;
void (*paint_ptr)(Display *dsp, Drawable drw, GC g);
void (*mouse_ptr)(int x, int y, int b, int s);

/* create_dockapp
 *	PARAMETERS:
 *		char * appname    : name of application to be registered with X
 *		char * geometry   : X geometry string to use when creating the window
 *		char **interface  : the XPM to use for the background and shape mask
 *		int withdrawn     : if withdrawn is > 0 then the app is drawn in
 *                          withdrawn state (WindowMaker DockApps)
 *                          if withdrawn is zero then it is drawn as a 
 *                          normal window
 *      int argc          : number of command line parameters
 *      char **argv       : array of command line parameters
 *	RETURN:
 *		int : returns non-zero on success
 */
int
create_dockapp(	char *appname,
				char *geometry,
				char **interface,
				int withdrawn,
                int argc,
                char **argv);

/* run_dockapp
 *	PARAMETERS:
 *		none
 *	RETURN:
 *		int : returns non-zero on success
 */
int
run_dockapp();

/* request_paint - requests that the paint function be called
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
request_paint();

/* clear_window - clears the draw window
 *  PARAMETERS:
 *      none
 *  RETURN:
 *      none
 */
void
clear_window();

/* set_paint_func
 *	PARAMETERS:
 *		void (*func)(Display *dsp, Drawable drw, GC g) : pointer to paint 
 *                                                      function
 *	RETURN:
 *		int : returns non-zero on success
 */
int
set_paint_func(	void (*func)(Display *dsp, Drawable drw, GC g));

/* set_mouse_func
 *	PARAMETERS:
 *		void (*func)(int x, int y, int b, int s) : pointer to mouse function 
 *	RETURN:
 *		int : returns non-zero on success
 */
int
set_mouse_func(	void (*func)(int x, int y, int b, int s));

/* get_display()
 *	RETURN:
 *		Display * : returns display
 */
Display*
get_display();

/* get_screen()
 *	RETURN:
 *		int : returns screen
 */
int
get_screen();

/* get_black_pixel()
 *	RETURN:
 *		unsigned long : black pixel value
 */
unsigned long
get_black_pixel();

/* get_white_pixel()
 *	RETURN:
 *		unsigned long : white pixel value
 */
unsigned long
get_white_pixel();


#endif /* not __DOCKAPP_H__ */
